package PureMessage::Enduser::Auth::Authenticator::LDAP;

###############################################################################
# Required inclusions.
###############################################################################
use strict;                         # Activate compile-time syntax checking
use Net::LDAP;                      # Needed for LDAP connectivity
use Net::LDAP::Constant qw(LDAP_PARTIAL_RESULTS LDAP_REFERRAL);
use URI::ldap;                      # Needed for LDAP URI parsing
use PureMessage::Template qw(expand_template);

###############################################################################
# Derive ourselves from our base class.
###############################################################################
use base qw( PureMessage::Enduser::Auth::Authenticator );

###############################################################################
# Version number.
###############################################################################
our $VERSION = '2.000';

###############################################################################
# Module configuration.
###############################################################################
our %module_config;

###############################################################################
# Allow for optional debugging of the data stream.
###############################################################################
my $DEBUG = $ENV{AUTH_DEBUG};

###############################################################################
# Helper methods to make logging to "httpd_error.log" a bit easier.
###############################################################################
sub log_message
{
    my ($msg) = @_;
    chomp( $msg );
    print STDERR "EU-LDAP-AUTH: $msg\n";
}
sub log_debug
{
    my ($msg) = @_;
    log_message( "[debug] $msg" ) if ($DEBUG);
}
sub log_warning
{
    my ($msg) = @_;
    log_message( "[warn] $msg" );
}
sub log_error
{
    my ($msg) = @_;
    log_message( "[error] $msg" );
}

###############################################################################
# Initialization.
###############################################################################
# Configuration directives:
#   ldap_server
#       - required
#       - list of LDAP servers to connect to for authentication
#   dn_discovery
#       - if non-zero, enables the auto-discovery of the User DN
#   bind_dn
#       - DN for bind to LDAP server
#       - if discovery is enabled, this pertains -only- to the discovery phase
#       - if discovery is disabled, this pertains to the actual authentication
#   bind_password/bind_pw
#       - password for bind to LDAP server
#       - only valid during discovery phase
#   base_dn
#       - required
#       - base DN under which all searches should be performed
#   filter
#       - required
#       - filter/search to be performed to find user record
#   attribute_mail
#       - name of LDAP attribute that is queried for the "primary e-mail
#         address" for the user
#       - defaults to "mail" if not provided
#   attribute_mail_index
#       - specifies the index to the result record whose "attribute_mail" we
#         should be returning as the primary e-mail address for this user
#       - defaults to "0" if not provided
#       - only useful if the LDAP server -always- returns multiple e-mail
#         attributes for users and we -always- know that we want the 'index'th
#         one
#   debug
#       - enables debug logging
###############################################################################
sub init
{
    %module_config = @_;

    ###########################################################################
    # SANITY CHECK: Make sure that our config has all of the required fields.
    ###########################################################################
    foreach my $required (qw( ldap_server base_dn filter ))
    {
        if ( (!$module_config{$required}) ||
             ($module_config{$required} =~ /^\s*$/)
           )
        {
            die "EUI LDAP Authenticator config is missing '$required'.\n" .
                "LDAP authentication disabled.\n";
        }
    }

    ###########################################################################
    # Set defaults for any values that are missing.
    ###########################################################################
    my %defaults = (
        'attribute_mail'        => 'mail',
        'attribute_mail_index'  => 0,
        'bind_pw'               => $module_config{'bind_password'},
        );
    foreach my $attr (keys %defaults)
    {
        if ((!$module_config{$attr}) || ($module_config{$attr} eq ''))
        {
            $module_config{$attr} = $defaults{$attr};
        }
    }

    ###########################################################################
    # Allow for debugging to be specified by the configuration as well as from
    # ENV{AUTH_DEBUG}
    ###########################################################################
    $DEBUG ||= $module_config{'debug'};

    ###########################################################################
    # Return a handle to the authenticator function.
    ###########################################################################
    return \&Authenticate;
}

###############################################################################
# LDAP Authentication.
###############################################################################
sub Authenticate
{
    my ($in_user, $in_password, %opt) = @_;

    ###########################################################################
    # Abort if we don't have everything.
    ###########################################################################
    return unless ($in_user and $in_password);

    ###########################################################################
    # Connect to the LDAP server, get the "distinguishedName" of the user that
    # we're attempting to authenticate, and authenticate as that user.
    ###########################################################################
    my ($primary);
    eval
    {
        #######################################################################
        # Get our config out of the module config...
        #######################################################################
        my $info = {
            'username'      => $in_user,
            'userid'        => $in_user,
            'password'      => $in_password,
            'server'        => $module_config{'ldap_server'},
            'bind_dn'       => $module_config{'bind_dn'},
            'bind_pw'       => $module_config{'bind_pw'},
            'base_dn'       => $module_config{'base_dn'},
            'filter'        => $module_config{'filter'},
            'attr'          => $module_config{'attribute_mail'},
            'attr_index'    => $module_config{'attribute_mail_index'},
            };

        #######################################################################
        # If the "username" was provided in a form that looks like an MS
        # "domain\userid", split it out into its component bits.
        #######################################################################
        if ($in_user =~ /^\s*(.+)\\+(.+)\s*$/)
        {
            $info->{'domain'} = $1;
            $info->{'userid'} = $2;
            log_debug( "Username looks like 'domain\\userid' format" );
            log_debug( "... domain: $info->{'domain'}" );
            log_debug( "... userid: $info->{'userid'}" );
        }

        #######################################################################
        # Expand out some of the config options.
        #######################################################################
        foreach my $field (qw( base_dn bind_dn filter ))
        {
            log_debug( "Expanding $field: $info->{$field}" );
            $info->{$field} = expand_template(
                                $info->{$field},
                                'vars' => $info,
                                );
            log_debug( "... expanded to: $info->{$field}" );
        }

        #######################################################################
        # Auto-discover the User DN.
        #######################################################################
        if ($module_config{'dn_discovery'})
        {
            $info = _discover_user_dn( $info );
        }

        #######################################################################
        # Authenticate the user.
        #######################################################################
        $info->{'bind_pw'} = $in_password;
        $info = _authenticate( $info );
        $primary = $info->{'email'};
    };
    if ($@)
    {
        log_error( $@ );
    }

    ###########################################################################
    # If we found a primary e-mail address, return that back to the caller.
    ###########################################################################
    return (email => $primary) if ($primary);
    log_error( "Didn't find e-mail account for user '$in_user'" );
    return;
}

###############################################################################
# "Discovers" the User DN by performing a search against the LDAP directory.
# If successful, this method returns a hash-ref of information back to the
# caller.  This method calls 'die()' on error.
###############################################################################
sub _discover_user_dn
{
    my ($info) = @_;
    log_debug( 'Discovering User DN ...' );

    ###########################################################################
    # Connect to the LDAP server.
    ###########################################################################
    log_debug( '... connecting to LDAP server' );
    my $ldap = Net::LDAP->new( $info->{'server'},
                               'version'    => 3,
                               'debug'      => ($DEBUG > 1 ? $DEBUG : 0),
                             );
    die "Unable to connect to LDAP server.\n" if (!$ldap);
    log_debug( '... ... connected' );

    ###########################################################################
    # Bind, following referrals
    ###########################################################################
    my $msg;
    if ($info->{'bind_dn'} && $info->{'bind_pw'})
    {
        log_debug( "... binding to: $info->{'bind_dn'}" );
        $msg = $ldap->bind( 'dn' => $info->{'bind_dn'},
                            'password' => $info->{'bind_pw'},
                          );
        if ( ($msg->code() == LDAP_PARTIAL_RESULTS) ||
             ($msg->code() == LDAP_REFERRAL)
           )
        {
            # Follow the LDAP referral, -WITHOUT- a filter (so we don't do the
            # query but instead just do a bind).
            log_debug( "... ... got LDAP bind referral response, following" );
            my $results = _follow_ldap_referral( $msg,
                            { 'bind_dn' => $info->{'bind_dn'},
                              'bind_pw' => $info->{'bind_pw'},
                            } );
            if (!$results)
            {
                die "Unable to bind; LDAP referrals exhausted.\n";
            }
            $msg  = $results->{'msg'};
            $ldap = $results->{'ldap'};
            $info->{'server'}  = $results->{'server'};
            $info->{'base_dn'} = $results->{'base_dn'};
        }
    }
    else
    {
        log_debug( "... binding anonymously" );
        $msg = $ldap->bind();
    }
    $msg->code() and die "Unable to bind to LDAP; " . $msg->error() . "\n";
    log_debug( "... ... bound" );

    ###########################################################################
    # Search for the User DN, following referrals.
    ###########################################################################
    my $attr = 'dn';
    log_debug( "... querying user attribute: $attr" );
    $msg = $ldap->search( 'base'    => $info->{'base_dn'},
                          'filter'  => $info->{'filter'},
                          'attrs'   => [$attr],
                        );
    if ( ($msg->code() == LDAP_PARTIAL_RESULTS) ||
         ($msg->code() == LDAP_REFERRAL)
       )
    {
        log_debug( "... ... got LDAP referral response, following" );
        my $results = _follow_ldap_referral( $msg,
                        { 'filter'  => $info->{'filter'},
                          'attr'    => $attr,
                        } );
        if (!$results)
        {
            die "Unable to query user record; LDAP referrals exhausted.\n";
        }
        $msg = $results->{'msg'};
        $info->{'server'}  = $results->{'server'};
        $info->{'base_dn'} = $results->{'base_dn'};
    }
    $msg->code() and die "Unable to query user record; " . $msg->error() . "\n";
    log_debug( "... ... query complete" );

    ###########################################################################
    # Extract the DN out of the queried result record.
    ###########################################################################
    my @entries = $msg->all_entries();
    if (scalar(@entries) == 0)
    {
        die "Unable to locate user record '$info->{'username'}'.\n";
    }
    elsif (scalar(@entries) > 1)
    {
        log_warning( "Filter returned multiple possible user records for '$info->{'username'}'" );
    }
    $info->{'bind_dn'} = $entries[0]->dn();
    log_debug( "... discovered User DN: $info->{'bind_dn'}" );

    ###########################################################################
    # Finished discovery; unbind and return results.
    ###########################################################################
    $ldap->unbind();
    return $info;
}

###############################################################################
# Authenticates the user based on the provided information.
###############################################################################
sub _authenticate
{
    my ($info) = @_;
    log_debug( "Authenticating user ..." );

    ###########################################################################
    # Connect to the LDAP server.
    ###########################################################################
    log_debug( '... connecting to LDAP server' );
    my $ldap = Net::LDAP->new( $info->{'server'},
                               'version'    => 3,
                               'debug'      => ($DEBUG > 1 ? $DEBUG : 0),
                             );
    die "Unable to connect to LDAP server.\n" if (!$ldap);
    log_debug( '... ... connected' );

    ###########################################################################
    # Bind (authenticate), following referrals
    ###########################################################################
    log_debug( "... binding to: $info->{'bind_dn'}" );
    my $msg = $ldap->bind( 'dn' => $info->{'bind_dn'},
                           'password' => $info->{'bind_pw'},
                         );
    if ( ($msg->code() == LDAP_PARTIAL_RESULTS) ||
         ($msg->code() == LDAP_REFERRAL)
       )
    {
        # Follow the LDAP referral, -WITHOUT- a filter (so we don't do the
        # query but instead just do a bind).
        log_debug( "... ... got LDAP bind referral response, following" );
        my $results = _follow_ldap_referral( $msg,
                        { 'bind_dn' => $info->{'bind_dn'},
                          'bind_pw' => $info->{'bind_pw'},
                        } );
        if (!$results)
        {
            die "Unable to bind; LDAP referrals exhausted.\n";
        }
        $msg  = $results->{'msg'};
        $ldap = $results->{'ldap'};
        $info->{'server'}  = $results->{'server'};
        $info->{'base_dn'} = $results->{'base_dn'};
    }
    $msg->code() and die "Unable to bind to LDAP; " . $msg->error() . "\n";
    log_debug( "... ... bound (authenticated)" );

    ###########################################################################
    # Search for the primary e-mail address, following referrals
    ###########################################################################
    log_debug( "... querying user attribute: $info->{'attr'}" );
    $msg = $ldap->search( 'base'    => $info->{'base_dn'},
                          'filter'  => $info->{'filter'},
                          'attrs'   => [$info->{'attr'}],
                        );
    if ( ($msg->code() == LDAP_PARTIAL_RESULTS) ||
         ($msg->code() == LDAP_REFERRAL)
       )
    {
        # Follow the referral, WITH the query, as we're looking for an actual
        # response.
        log_debug( "... ... got LDAP referral response, following" );
        my $results = _follow_ldap_referral( $msg,
                        { 'filter'  => $info->{'filter'},
                          'attr'    => $info->{'attr'},
                          'bind_dn' => $info->{'bind_dn'},
                          'bind_pw' => $info->{'bind_pw'},
                        } );
        if (!$results)
        {
            die "Unable to query user record; LDAP referrals exhausted.\n";
        }
        $msg  = $results->{'msg'};
        $ldap = $results->{'ldap'};
        $info->{'server'}  = $results->{'server'};
        $info->{'base_dn'} = $results->{'base_dn'};
    }
    $msg->code() and die "Unable to query user record; " . $msg->error() . "\n";
    log_debug( "... ... query complete" );

    ###########################################################################
    # Extract primary e-mail address out of the queried result record.
    ###########################################################################
    my @entries = $msg->all_entries();
    if (scalar(@entries) > 0)
    {
        my $index   = $info->{'attr_index'};
        log_debug( "... ... retrieving attribute value at index $index" );

        my $value   = $entries[$index]->get_value( $info->{'attr'} );
        if ($value)
        {
            $info->{'email'} = $value;
            log_debug( "... found primary e-mail: $value" );
        }
    }

    ###########################################################################
    # Finished authentication; unbind and return results.
    ###########################################################################
    $ldap->unbind();
    return $info;
}

###############################################################################
# Follows an LDAP referral response (and all referrals it may lead us to).
# Referrals are followed in a "breadth-first" fashion, and the first referral
# to provide us with a suitable response is accepted.
###############################################################################
sub _follow_ldap_referral
{
    my ($msgin, $opts) = @_;

    ###########################################################################
    # Get our config items out of the provided options.
    ###########################################################################
    my $filter = $opts->{'filter'};
    my $attr   = $opts->{'attr'};
    my $binddn = $opts->{'bind_dn'};
    my $bindpw = $opts->{'bind_pw'};

    ###########################################################################
    # Keep track of what referrals we've already followed, to try to actively
    # avoid potential infinite query loops.
    ###########################################################################
    my %followed = ();

    ###########################################################################
    # Attempt to process each of the LDAP referrals we were given.
    ###########################################################################
    my @referrals = $msgin->referrals();
    while (my $ref = shift @referrals)
    {
        log_debug( "... ... LDAP referral: $ref" );

        #######################################################################
        # Check to make sure that we haven't followed this referral before.
        #######################################################################
        if (exists $followed{$ref})
        {
            log_warning( "Potential recursive LDAP referral: $ref" );
            next;
        }
        $followed{$ref} ++;

        #######################################################################
        # Extract the information we need out of the LDAP referral, and rebuild
        # a server connection string.
        #######################################################################
        my $uri     = URI::ldap->new( $ref );
        my $scheme  = $uri->scheme();
        my $host    = $uri->host();
        my $port    = $uri->port();
        my $basedn  = $uri->dn();

        my $server = "$scheme://$host";
        $server .= ":$port" if ($port);

        #######################################################################
        # Attempt to follow the referral...
        #######################################################################
        my ($ldap, $response);
        eval
        {
            ###################################################################
            # Connect
            ###################################################################
            log_debug( "... ... ... connecting" );
            $ldap = Net::LDAP->new( $server,
                                    'version'    => 3,
                                    'debug'      => ($DEBUG > 1 ? $DEBUG : 0),
                                  );
            die "unable to connect to '$server'\n" if (!$ldap);
            log_debug( "... ... ... ... connected" );

            ###################################################################
            # Bind
            ###################################################################
            my $msg;
            if ($binddn && $bindpw)
            {
                log_debug( "... ... ... binding with credentials" );
                $msg = $ldap->bind( 'dn' => $binddn, 'password' => $bindpw );
            }
            else
            {
                log_debug( "... ... ... binding anonymously" );
                $msg = $ldap->bind();
            }
            if ( ($msg->code() == LDAP_PARTIAL_RESULTS) ||
                 ($msg->code() == LDAP_REFERRAL)
               )
            {
                # Add other referrals to our list of ones to follow
                push( @referrals, $msg->referrals() );
                die "got bind referrals\n";
            }
            elsif ($msg->code())
            {
                # Choke if the bind failed
                die "unable to bind; " . $msg->error() . "\n";
            }
            log_debug( "... ... ... ... bound" );
            $response = $msg;

            ###################################################################
            # Query, if we were given a filter.
            ###################################################################
            if ($filter)
            {
                undef $response;
                log_debug( "... ... ... querying" );
                $msg = $ldap->search( 'base'    => $basedn,
                                      'filter'  => $filter,
                                      'attrs'   => [$attr],
                                    );
                if ( ($msg->code() == LDAP_PARTIAL_RESULTS) ||
                     ($msg->code() == LDAP_REFERRAL)
                   )
                {
                    # Add other referrals to our list of ones to follow
                    push( @referrals, $msg->referrals() );
                    die "got query referrals\n";
                }
                elsif ($msg->code())
                {
                    # Choke if the query failed
                    die "query failed; " . $msg->error() . "\n";
                }
                $response = $msg;
            }
        };
        if ($@)
        {
            log_debug( "... ... ... ... $@" );
        }

        #######################################################################
        # Go onto the next LDAP referral if we don't have a suitable response
        # yet.
        #######################################################################
        next if (!$response);

        #######################################################################
        # IF we managed to get a suitable response from following this
        # referral, return it back to the caller.
        #######################################################################
        my $results = {
            'server'    => $server,
            'ldap'      => $ldap,
            'msg'       => $response,
            'base_dn'   => $basedn,
            };
        return $results;
    }

    ###########################################################################
    # Didn't find a match; return empty-handed.
    ###########################################################################
    log_debug( "... ... LDAP referrals exhausted" );
    return;
}

###############################################################################
1;
