package PureMessage::Enduser::User;
use strict;
use warnings;
use Storable qw(freeze thaw);
use PureMessage::Enduser::Config qw( %config);
use PureMessage::Manager::Lang qw(mtext);
use PureMessage::List;
use PureMessage::Map;
use PureMessage::ListFile;
use PureMessage::Quarantine;
use PureMessage::Enduser::Auth::Email;
use PureMessage::Enduser::Auth qw(Authorize);
use PureMessage::ListManager;
use PureMessage::Logger;
use PureMessage::DataStore;
use Date::Parse;
use Time::Local qw(timegm);
use POSIX qw(strftime);
use PureMessage::Enduser::Session;
use PureMessage::Preferences;
use PureMessage::Resources::Manager;
use PureMessage::QMan qw(prepare_msg_preview);

our $RESOURCE_MANAGER = undef;
our $RESOURCE_MANAGER_OWNER = undef;

use Mail::Address;

our $DEBUG = $ENV{PMX_ENDUSER_DEBUG};
our $USER_LIST;

my @QUARANTINE_FIELDS = qw( m_date m_enduser m_reason h_from h_to m_date c_subject_utf8 m_spam_probability);
my @EXTRA_QUARANTINE_FIELDS = qw( qid );

my %QUARANTINE_FIELDS_DATA=( m_enduser          => { sort => "s"},
                             c_reason_csv       => { sort => "s"},
                             h_from             => { sort => "s"},
                             h_to               => { sort => "s"},
                             m_date             => { sort => 'n'},
                             h_subject          => { sort => 's'},
                             m_spam_probability => { sort => 'n'},
                             qid                => { sort => 'n'},
                           );

# Enduser user operations. A user maps directly to an email@address
# encapsulates all the things an enduser might want to do into one
# object

sub _resource_manager {
   return $RESOURCE_MANAGER if $RESOURCE_MANAGER && $RESOURCE_MANAGER_OWNER eq $$;
   $RESOURCE_MANAGER = PureMessage::Resources::Manager->new( writable => 1);
   $RESOURCE_MANAGER_OWNER = $$;
   return $RESOURCE_MANAGER;
}


=item new ($email_address)

Returns a user object

=cut

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $email = shift;
    my %opts = @_;
    my %self = ();
    return 0 unless $email;
    $self{email} = lc($email);
    $self{logger} = $opts{logger} || PureMessage::Logger->new(ident => "endusr_rpc");
    $self{limit} = $opts{limit} || $config{max_view_messages};
    return bless( \%self, $class);
}

=item authorize($identity, $pass)

Will attempt to verify the passed information using the currently 
configured authentication backend if this user is a valid user as
defined in the C<allow.users> conf setting. If successful an object
will be returned, otherwise will die with the appropriate error.

If email address is needed, it can be accessed using the C<email>.
Any extra information returned by the authentication backend can
be accessed in the C<auth> key.
method.

=cut


sub authorize {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my ($identity, $pass,%opts) = @_;
  my %self = ();
  die mtext("authorize_invalid_identity",$identity) . "\n" unless $identity;
  $self{logger} = $opts{logger} || PureMessage::Logger->new(ident => "eui_local");
  if (my %info = Authorize($identity,$pass,logger => $self{logger})) { 

    unless( valid_user(lc($info{email})) ) {
      $self{logger}->log("INFO", "user '$self{email}' has an email address that doesn't match a valid user as defined by the list '$config{allow}->{users}'") if $self{logger};
      die mtext("authorize_invalid_identity",$identity) . "\n";
    }
    $self{email} = lc($info{email});
    $self{auth} = \%info;
  } else {
    return; # nope nope bad password
  }
   
  return bless( \%self, $class);
}

=item session

Returns the users last know session object or undef if none.

=cut

sub session {
  my $self = shift;
  my %options = @_;
  $self->_data_store();
  my $store = $self->{ds}->get($self->{email});
  my $session_id = $store->retrieve('session_id');
  my $session;
  if ($session_id) {
    $session = PureMessage::Enduser::Session->restore($session_id);
  }
  
  unless($session) {
    if ($options{create}) {
      $session = PureMessage::Enduser::Session->new();
      return unless $session;
      $store->store('session_id',$session->id());
    } 
  }
  return $session;
}

=item store
 
 Have the user object save it's state to it's persistent data store.

=cut
sub _data_store {
  my $self = shift;
  unless($self->{ds}) {
    $self->{ds} = _get_resource('enduser_user_store');
  }
  $self->{ds};
}


sub store {
  my $self = shift;
  $self->_data_store();
  my $store = $self->{ds}->get($self->{email});
  my $state = $self->state();
  # only save our state if we have query data
  return unless defined($state->{query_valid_till}) 
            and $state->{query_valid_till} > time;

  my $view_data = delete($state->{view});
  $store->store('user_state',freeze($state));
  $store->store('user_state_view_data',freeze($view_data))
    if !$self->{must_restore_view_data};
}

=item restore

  Have the user object restore it's state from it's persistent
  data store.  Note that if it has been longer than 
  C<$config{session}->{args}->{expire}->{user_data}> since this
  data was saved it will be discarded instead and the quarantine
  queried anew.

=cut

sub restore {
  my $self = shift;
  $self->_data_store();
  my $store = $self->{ds}->get($self->{email});
  my $data = $store->retrieve('user_state');
  my $state = $data ? thaw($data) : undef;
  
  if ($state) {
      
    if($state->{query_valid_till} > time) {
      $self->state(thaw($data)) if $data;
      $self->{must_restore_view_data} = 1;
    } else {
      # the state has expired so delete whatever we've got
      $store->delete('user_state');
      $store->delete('user_state_view_data');
    }
  }

}

sub _restore_view_data {
 my $self = shift;
 return unless $self->{must_restore_view_data};
 $self->_data_store();
 my $store = $self->{ds}->get($self->{email});
 my $data = $store->retrieve('user_state_view_data');
 my $state = $self->state();
 $state->{view} = thaw($data) if $data;
 $self->state($state) if $data;
 $self->{must_restore_view_data} = 0;
}


=item valid_user($username)

Returns true if this user a user that is allowed to connect as defined
by the $config{allow}->{users} list.

=cut

sub valid_user {
  my $user = shift;
  return unless $config{allow}->{users}; # by default allow none
  unless ($USER_LIST) {
    eval { $USER_LIST = _get_resource($config{allow}->{users}); };
    unless ($USER_LIST) {
        $USER_LIST = PureMessage::List->new($config{allow}->{users});
    }
  }
  die mtext('user_invalid_access_list',$config{allow}->{users}) unless $USER_LIST;
  return $USER_LIST->match($user); 
}


=item email()

Method that returns the users email address.

=cut

sub email {
  my $self = shift;
  return $self->{email};
}

=item send_notification()

Attempts to que an authorization message for this user.

=cut

sub send_notification {
  my $self = shift;
  unless(valid_user($self->{email})) {
    $self->{logger} = PureMessage::Logger->new(ident => "eui_local") unless $self->{logger};
    $self->{logger}->log("INFO", "user '$self->{email}' has an email address that doesn't match a valid user as defined by the list '$config{allow}->{users}'") if $self->{logger};
    return;
  }
  return PureMessage::Enduser::Auth::Email::EmailKey($self->{email}) if $self->{email};
  return;
}


# this maps the qview to the sub that is actually responsible
# for generating the mset object
my %VIEWS = ( visible => q((not ( any m_enduser = 'delete %s' or any m_enduser = 'approve %s'))),
              'delete' => q(any m_enduser = 'delete %s'),
              approve => q(any m_enduser = 'approve %s'),
              all => undef,
            );


=item available_qviews()

Returns the possible qviews that can be queried. Currently
there is only: hidden, visible

=cut

sub available_qviews {
  my $self = shift;
  return sort keys %VIEWS;
}

# internal method gets the passed email address quarantine messages
sub _get_user_view {
  my ($self,$view) = @_;
  my $mstore = PureMessage::Quarantine->new;
  my %opts;
  my @where;
  my @reasons = ref($config{allow}->{reasons}) eq 'ARRAY' ? 
    @{$config{allow}->{reasons}} : ($config{allow}->{reasons});
  return $mstore->get_user_views($self->{email}, { name => $view, reasons => \@reasons , limit => $self->{limit}});
}

=item user_view($view,@emails)

Returns a view object that will return the quarantine messages
that match the passed view for the passed emails.

=cut

sub user_view {
  my ($self,$view) = @_;
  return unless exists($VIEWS{$view});
  return $self->_get_user_view($view);
}

=item is_query_too_soon([@views]) 

Will return an error message if this user cannot query
the quarantine because of an administrative time limit,
otherwise will return undef.

=cut

sub is_query_too_soon {
  my ($self,@views) = @_;
  @views = $self->available_qviews() unless $#views >= 0;
  my $must_wait;
  my $time = time;
  foreach my $view (@views) {
      # query rate limiting, if too soon then return an error. This assumes
      # that we are using persistant sessions. If that's not true this won't 
      # work.
      $must_wait = defined($self->{state}->{limit}->{$view}) ? $self->{state}->{limit}->{$view} - $time : 0; 
      return mtext('query_too_soon',$config{limit}->{query_wait},$must_wait)
        if $must_wait > 0; 
  } 
}

=item initialize_qview($view)

Attempt to initialize quararantine query for the requested view. Possible
views are: visible, hidden

=cut


sub initialize_qview {
  my ($self,@views) = @_;
  my (%state,@view_data,%temp);
  @views = ('visible','delete') unless $#views >= 0;
  #@views = $self->available_qviews() unless $#views >= 0;
  foreach my $view (@views) {
      
      if (my $err = $self->is_query_too_soon(@views)) {
        die $err . "\n"
      }
      my $vw = $self->user_view($view);
      next unless $vw; # TODO this is probably an error
      
      my $i = 0;

      while(my $row = $vw->fetch) {
	 my @fields = $row->fields;
	 my %data;
	 $data{qid} = $row->message_id;
	 foreach my $f (@QUARANTINE_FIELDS) {
	     $data{$f} = $row->field($f);
	 }
	 push(@{$state{$view}},\%data);
	 $i++;
      }
      push(@view_data,{ view => $view, row_count => $i });    
  }
  foreach my $view (@views) {
    $self->{state}->{limit}->{$view} = $config{limit}->{query_wait} ? time + $config{limit}->{query_wait} : 0;
    $self->{state}->{view}->{$view} = $state{$view};
    $self->{state}->{order}->{$view} = ''; # TODO have a default sort and pop this field with it
  }
  
  if (ref($self->{state}->{view_data}) eq 'ARRAY') {

    foreach my $new_view (@view_data) {
        
        my $found = 0;
        foreach my $existing_view (@{$self->{state}->{view_data}}) {

          if($existing_view->{view} eq $new_view->{view}) {
              $existing_view->{row_count} = $new_view->{row_count};
              $found =1;
          }

        }
        push(@{$self->{state}->{view_data}},$new_view) unless $found;
    }
    
  } else {
    $self->{state}->{view_data} = \@view_data;
  }
  $self->{must_restore_view_data} = 0;
  $self->{state}->{query_valid_till} = time + $config{session}->{args}->{expire}->{user_data};
  return @view_data;
}

=item get_qview(@view)

Returns the same data as init_qview without querying the quarantine to
get it.  If you need the count of rows in a view use this method.

=cut

sub get_qview {
  my ($self,@views) = @_;
  my (%state,@view_data,%temp);
  $self->_restore_view_data();
  @views = ('visible','delete') unless $#views >= 0;
  foreach my $view (@views) {
    my @view = grep { $_->{view} eq $view } @{$self->{state}->{view_data}};
    # if this returns undef then it is expired
    push(@view,$self->initialize_qview($view)) unless scalar(@view) and $view[0]->{view} eq $view;
    push(@view_data,@view);
  } 
  return @view_data;
}


=item get_qrows($view,$start,$end,@sort)

Will attempt to get the C<get_msg_meta> information for each of the
rows in the $view between and including $start and $end. Rows are
numbered from 0.

It is possible for less rows to be returned if they don't exist.

If initialize_qview hasn't been called, it will be called automatically
for you.

@sort is the name of the fields to sort by, prefix with - to sort
desc

=cut

sub format_today_date {
  my $date = shift;
  # str2time returns you a GMT time
  
  # calculate when the start of today is
  my @time =gmtime();
  $time[$_] = 0 foreach qw(0 1 2);
  my $today_starts_at_gmt =  timegm(@time);
  
  my $utime = str2time($date,'GMT');
  return $date unless $utime;
  my @time_pieces = gmtime($utime + $config{gmt_offset});
  if ($utime >= $today_starts_at_gmt) {
    #return a today formatted date
    return strftime('%l:%M %p',@time_pieces);
  } else {
    return strftime('%c',@time_pieces);
  }
 

}

sub get_qrows {
  my ($self,$view,$start,$end,@sort) = @_;
  my @result;
  $self->_restore_view_data();
  $self->initialize_qview($view) unless exists($self->{state}->{view}->{$view});
  my $data = $self->{state}->{view}->{$view};
  my $last = $#$data;
  
  # make sure these numbers are sane
  $end = $last if $end < 0 or $end > $last;
  $start = 0 if $start < 0;
  unless( $self->{state}->{order}->{$view} eq join('|',@sort)) {
      # only sort if this isn't how it's currently sorted 
      if (my @date = grep(/date/i,@sort)) {
        # We don't want to reparse these every time so we'll just parse and cache
        # rows and keys as we need them
        require Date::Parse;
        while(my $field = pop(@date)) {
            $field =~ s/^-//;
            $field =~ s/(:.{1,30})$//; # remove any decorators
            for(my $index = 0; $index <= $#$data; $index++) {
               next if defined $data->[$index]->{"$field!parsed"}; # it's cached
               my $utime = Date::Parse::str2time($data->[$index]->{$field});
               # if not utime then we couldn't parse it.. 
               $utime = 0 unless $utime; # will be undef on err, so lets sort them at the beginning
               $data->[$index]->{"$field!parsed"} = $utime;
            }
        }
      }

      # do a special case for h_from:addr which doesn't really exist but means
      # we parse and sort on email address
      if (my @fields = grep(/^.+:addr$/i,@sort)) {
        # We don't want to reparse these every time so we'll just parse and cache
        # rows and keys as we need them
        require Mail::Address;
        while(my $field = pop(@fields)) {
            $field =~ s/^-//; # we don't at this stage care about asc/desc
            $field =~ s/:addr$//;
            for(my $index = 0; $index <= $#$data; $index++) {
               next if defined $data->[$index]->{"$field!parsed"}; # it's cached
               my @addrs = Mail::Address->parse($data->[$index]->{$field});
               my $addr_text = '';

               if ($#addrs != -1) {
                $addr_text = join(', ',$_->address()) foreach @addrs;
               }
               
               $data->[$index]->{"$field!parsed"} = $addr_text;
               $data->[$index]->{"$field:addr"} = $addr_text;

            }
        }
      }
 
      if (my @fields = grep(/^.+:tzt$/i,@sort)) {
        # We don't want to reparse these every time so we'll just parse and cache
        # rows and keys as we need them
        while(my $field = pop(@fields)) {
            $field =~ s/^-//; # we don't at this stage care about asc/desc
            $field =~ s/:tzt$//;
            for(my $index = 0; $index <= $#$data; $index++) {
               next if defined $data->[$index]->{"$field:tzt"}; # it's cached
               $data->[$index]->{"$field:tzt"} = format_today_date($data->[$index]->{$field});

            }
        }
      }

      if (@sort) {
        $self->{state}->{order}->{$view} = join('|',@sort); # rememembed this sort 
        while(my $field = pop(@sort)) {
          my $desc = $field =~ s/^-//;
          my $parsed = $field =~ s/:addr$//; # it has a :addr appended to it? 
          next unless grep(/^$field$/,(@QUARANTINE_FIELDS,@EXTRA_QUARANTINE_FIELDS));
          my $numeric = (exists($QUARANTINE_FIELDS_DATA{$field}) and $QUARANTINE_FIELDS_DATA{$field}->{'sort'}) eq 'n' ? 1 : 0;
          $field = "$field!parsed" if $field =~ /date/ or $parsed; # it is a date field
          if ($desc) {
             @{$data} = sort { $numeric ? $b->{$field} <=> $a->{$field} 
                                        : lc($b->{$field}) cmp lc($a->{$field})
                             } @{$data};
          } else {
             @{$data} = sort { $numeric ? $a->{$field} <=> $b->{$field} 
                                        : lc($a->{$field}) cmp lc($b->{$field})
                             } @{$data};
          }
        }
      }
 }

 @result = @{$data}[$start..$end];
 return \@result if scalar(@result);
 return;
}

=item get_msg_meta(@ids)
 
Retrieved the message meta information for the passed ids. The ids must have
been retrieved in a previous quarantine operation with this user as only the
message cache will be checked for information.

id == qid

=cut

sub get_msg_meta {
  my ($self,@msg_ids) = @_;
  $self->_restore_view_data(); 
  $self->initialize_qview() unless exists($self->{state}->{view_data}) and ref($self->{state}->{view_data}) eq 'ARRAY';
  
  my @found;
  OUTER: foreach my $id (@msg_ids) {
     foreach my $view (keys %{$self->{state}->{view}}) {
       foreach my $row ( @{$self->{state}->{view}->{$view}} ) {
          if ($row->{qid} eq $id) {
            push(@found,$row);
            next OUTER;
          }
       }
     }
  }
  
  return \@found;

}

=item get_msg_body($qid)

Will return the messages body content as a text string.  The length of the data
 returned is configurable by the admin.

=cut

sub get_msg_body {
  my ($self,$qid) = @_;
  my $data;
  my $mstore = PureMessage::Quarantine->new();
  my $message = $mstore->message($qid);
  return unless $message;
  # TODO make the length of returned data configurable
  # prepare_msg_preview() is part of on-demand meta-indexing.
  prepare_msg_preview($message);
  my $bdata = $message->cleaned_body;
  return '' unless defined $bdata;
  return $bdata;
}

=item get_msg($qid)

Will return the messages data (including headers) text string.  The length of the data
 returned is configurable by the admin.

=cut

sub get_msg {
  my ($self,$qid) = @_;
  return $self->get_msg_body($qid,1);
}


=item approve_msg(@qid)

Will release the message from the quarantine and mark it so that it doesn't appear
in subsequent quarantine views

=cut

use PureMessage::QMan qw(approve trash);

sub approve_msg {
  my ($self,@qids) = @_;
  $self->_restore_view_data();
  return $self->_message_traversal("approve", \@qids);
  #return $self->_message_traversal('approve',\@qids,exec => \&approve,mark=>1);
}

=item delete_msg(@qid)

Will mark the message so that it doesn't appear in subsequent quarantine views

=cut

sub delete_msg {
  my ($self,@qids) = @_;
  $self->_restore_view_data();
  return $self->_message_traversal("delete", \@qids);
  #return $self->_message_traversal('delete',\@qids,result => 1,mark=>1);
}

=item undelete_msg(@qid)

Will unmark the message so that it reappear in subsequent quarantine views

=cut

sub undelete_msg {
  my ($self,@qids) = @_;
  $self->_restore_view_data();
  return $self->_message_traversal("undelete", \@qids);
  #return $self->_message_traversal('undelete',\@qids,result => 1,mark=>0,move_to=>'visible');
}


=item _message_traversal($enduser_operation,&sub,\@qids,%options)

For each passed qid will create a Quarantine message object
and execute &sub passing the object to it. Each message
has $enduser_operation added to it's enduser minfo field.

=cut

sub _message_traversal {
    my ($self, $action, $qids) = @_;

    my $logger = $self->{logger};
    my $q = PureMessage::Quarantine->new();
    $q->message_writable(1);

    # Get the message objects to work on
    my %result;
    my $map;
    # XXXX TODO the below call can fail silently because the 
    # $@ gets magically cleared somewhere on it's way back to
    # the package that called us.
    $map = _get_resource('recipient-aliases');
    # Even though the QMan approve() and trash() methods can accept 
    # multiple messages to process in one call, all those messages need to 
    # either be all going to a single address or to all of the default evelope 
    # recipients. We can't use this feature as we need to customize the 
    # recipeints on a per-message basis so we call each method once for each
    # message.
    foreach my $id (@$qids) {
	$logger->log("DEBUG", "get message from id: $id");
	my $msg = $q->message($id);
	unless ($msg) {
	    $logger->log("DEBUG", "Failed to find message $id");
	    next;
	}
	
	# This step shouldn't be required as envelope addresses in the 
	# quarantine are _suppoed_ to be valid but it's best to be defensive.
	my @to_cleaned;
	foreach my $email ($msg->minfo_value('Envelope-To')) {
	    my @temp = Mail::Address->parse($email);
	    next unless @temp;
	    push(@to_cleaned, map { $_->address } @temp);
	}

	# Of the recipients for this message, find out which ones we manage
	# and perform this action for each of them.
	my @managed_recips;
	foreach my $email (@to_cleaned) {
	    if ($map and lc($map->transform($email)) eq $self->{email}) {
		push(@managed_recips, lc($email));
	    }
	}
        $logger->log('INFO',"Attempting to $action message $id with recips: " . join(', ',@managed_recips));
	unless (@managed_recips) {
	    $logger->log("INFO", 
		"Failed to $action message $id, none of the recipients map to $self->{email}");
	    next;
	}

	# Approve, delete or undelete the message
	if ($action eq "approve") {
            # we need a logger for pmx_log so that reports can "see" this
            # release. We'll still log our "we done it" to the EUWI log on
            # success
            my $pmx_logger = PureMessage::Logger->new(ident => "quarantine");
	    approve($msg, 
		to        => \@managed_recips, 
                quarantine => $q,
                mark_user => $self->{email},
		logger    => $pmx_logger, 
		on_sent   => sub {
			my ($id, $msg, $to) = @_;
			$self->_move_cached_message("approve", $id);
			$result{$id}++;
			$logger->log("INFO", "Marked message $id as approved");
		});
	}
	elsif ($action eq "delete") {
	    trash($msg, 
		to         => \@managed_recips, 
		mark_user  => $self->{mail},
		logger     => $logger,
		on_trashed => sub {
			my $msg = shift;
			$self->_move_cached_message("delete", $msg->id);
			$result{$msg->id}++;
			$logger->log("DEBUG", "Marked message $id as deleted");
		});
	}
	elsif ($action eq "undelete") {

	    # XXX QMan doen't have a concept of undeleting a message. This
	    # workaround only works for messages that are partially trashed
	    # (marked as deleted by a user).If a message is fully trashed 
	    # (moved to the trash folder) by an admin, this won't bring it 
	    # back (the message may not even be found).
	    my $negate = 0;
	    my $refresh_aliases = 1;
	    my $action = eval{
                $msg->user_actions( $self->{email}, "new",
                                    $negate, $refresh_aliases)
            };
	    my ($msg, $level);
	    if (!$action || $action ne "new" || $@) {
		$@ = '?' unless $@;
		$msg = "User $self->{email} failed to remove message $id from the trash: $@";
		$level = "ERROR";
		warn "ERROR: $msg";
	    }
	    else {
		$msg = "User $self->{email} removed message $id from the trash";
		$level = "INFO";
		$self->_move_cached_message("visible", $id);
		$result{$id}++;
	    }
	    $logger->log($level, $msg);
	}
    }

    return \%result;
}

=item _remove_cached_message(@qid)

Removes all references to the passed quid's from this objects
internal cache.

=cut

sub _remove_cached_message {
  my ($self,@qid) = @_;
  $self->_restore_view_data();
  return unless defined($self->{state}->{view});
  foreach my $qid (@qid) {
     
     VIEW: foreach my $view (keys %{$self->{state}->{view}}) {
        my $data = $self->{state}->{view}->{$view};
        next unless $data;
        for(my $index = 0; $index <= $#{@$data}; $index++) {
          # remove the row
          if ($data->[$index]->{qid} == $qid) {
              splice(@$data,$index,1);
              next VIEW;
          }
        }
     }
  }
}

=item _move_cached_message($new_location,@qid)

Moves the mesages that match the passed quid's from the cache view they
currently reside in to the passed C<$new_location>.

=cut

sub _move_cached_message {
  my ($self,$new_location,@qid) = @_;
  $self->_restore_view_data();
  return unless defined($self->{state}->{view});
  QID: foreach my $qid (@qid) {
     
     VIEW: foreach my $view (keys %{$self->{state}->{view}}) {
        my $data = $self->{state}->{view}->{$view};
        next unless $data;
        next if $view eq 'all'; # we shouldn't move stuff around in the all view
        for(my $index = 0; $index <= $#$data; $index++) {
          # move the row
          if ($data->[$index]->{qid} eq $qid) {
              my $msg = splice(@$data,$index,1);
              push(@{$self->{state}->{view}->{$new_location}},$msg);
              $self->{state}->{order}->{view} = undef; # mark as unsorted
              my @r;
              
              # decrement count for where it used to be
              foreach my $r (@{$self->{state}->{view_data}}) {
                  $r->{row_count}-- if $r->{view} eq $view;
              }
              # add it into where it is now
              foreach my $r (@{$self->{state}->{view_data}}) {
                  $r->{row_count}++ if $r->{view} eq $new_location;
              }
              next QID;
          }
        }
     }
  }

}


=item available_attributes

Returns an array of user attributes that can be manipulated with calls
to C<attribute>.

=cut

my %ATTRIBUTES = ( vacation_till => \&vacation_till,
                   spam_optout => \&spam_optout,
                   digest_optout => \&digest_optout,
                   language      => \&language,
                   );


sub _get_resource {
   my $resource = shift;
   my $res;
   
   #
   # The recipient-aliases map is _very_ slow when using the database.
   # Allow the user to override the default behaviour and always use
   # the local resource. For this to be effective, the map should be
   # stored in a CDB file. To override the default (defined in
   # PureMessage::Enduser::Config::Validator), the following line
   # needs to be added to /etc/enduser/enduser.conf:
   #
   #     recipient_aliases_map_location=local
   #
   # Also, recipient-aliases.conf should be configured to use a cdb map.
   if ($resource && ($resource eq 'recipient-aliases')
           && ($config{'recipient_aliases_map_location'} eq 'local')) {
       return PureMessage::Map->new($resource);
   }

   eval {
       my $rm = _resource_manager();
       
       eval {
         $res = $rm->find($resource);
       };
       die "Failed to find the resource '$resource':$@\n" if $@;
       
       unless($res) {

          eval {
            $res = $rm->add($resource);
          };
          die "Failed to find or add the resource '$resource':$@\n" if $@;
       }

       die "Unable to create '$resource' resource\n" unless $res;
   };
   
   if(!$res or $@) {
       
     die "Error Initializing '$resource': $@";
     
   } else {
       
     return $res->native();
   
   }
}

# tell the resource manager to not hold on to any references of this object

sub _forget_resource {
   my $resource = shift;
   my $rm = _resource_manager();
       
   $rm->forget($resource);
}

sub vacation_till{
  my ($email,$op,$attrib,$value) = @_;
  # we are expecting an iso formatted date: YYYY-MM-DD 
  if ($op eq 'set') {
    die "Date '$value' is not a valid ISO date.\n" unless $value =~ /^\d{4}-[0-1]?\d-[0-3]?\d$/;
    my $time_value = str2time($value);
    
    if ($time_value > time + $config{limit}->{vacation_expiry}) {
      my $days = int($config{limit}->{vacation_expiry}/(60*60*24));
      die "Date '$value' exceeds the maximum hold days of $days you are allowed to set\n";
    } else {
      my $prefs = _get_resource('enduser-preferences');
      $prefs->expire_delay_iso($email,$value);
      $prefs->save();
      _forget_resource('enduser-preferences');
    }
  } else {
    my $prefs = _get_resource('enduser-preferences');
    my $expire = $prefs->expire_delay_iso($email);
    _forget_resource('enduser-preferences');
    return $expire;
  }

}



sub boolean_value {
  my $val = shift;
  return 1 if $val =~ /^1|yes|true$/i; 
  return 0;
}

# $listname,[op => value],%list_opts
# eg. 'whitelist',[add => 'me@here.com'],prepend => 1


sub _list_manip_res {
  my ($list_res,$ops,%list_opts) = @_;
  $list_opts{create} = 1;
  my $temp = $list_res;
  my $list_name = $list_res->name();
  my $ld = $temp->data();
  if ($ld->is_readonly ) {
    die "List is readonly '" . $list_res->name() . "'";
    return 0;
  }

  my $value;
  my @values;
  my $op = shift(@$ops); 
  my $n = sub {
          $value = shift(@$ops);
          @values = ref($value) eq 'ARRAY' ? (@$value) : ($value);
          print STDERR "_list_manip($list_name,$op,[" . join(',',@values) . "])\n" if $DEBUG;
      };

  while($op) {    
      
      if($op eq 'delete') {
         &$n;
         $ld->remove_items(@values);
      } elsif ($op eq 'clear') {
         $ld->clear();
         $ld->commit();
      } elsif($op eq 'add') {
         &$n;
         $ld->add_items(@values);
      } elsif($op eq 'exists') {
         my $found = 0;
         &$n;
         $ld->enumerate_items( sub { my $item = shift; $found++ if grep { $_ eq $item } @values;});
         return $found;
      } elsif($op eq 'get') {
         # we will return the data as an array ref
         my @data;
         $ld->enumerate_items( sub { my $item = shift; push(@data,$item)});
         return \@data;
      }else {
        die "Unknown list operation called: $op";
      }
      $op = shift(@$ops);
  }
  $ld->commit();
  $list_res->refresh();
}

sub spam_optout{
   my ($email,$op,$attrib,$boolean) = @_;
   print "SetAttribute($email,$op,$attrib,$boolean)\n" if $DEBUG;
   my $list_name = 'anti-spam-optouts';
   my $list = _get_resource($list_name);

   if ($op eq 'set') {
       if (boolean_value($boolean)) {
            _list_manip_res($list, ['add' => $email,'delete'=>"!$email"], prepend => 1)
                unless (_list_manip_res($list,['exists' => $email]));
       } 
       else {
           _list_manip_res($list, ['add' => "!$email",'delete'=> $email], prepend => 1)
                unless (_list_manip_res($list,['exists' => "!$email"]));
       }
     return 1;
   } 
   else {
       $list->refresh();
       return '1' if $list->match($email);
       return '0';
   }
}

sub digest_optout{
   my ($email,$op,$attrib,$boolean) = @_;
   print "SetAttribute($email,$op,$attrib,$boolean)\n" if $DEBUG;
   $boolean = 0 unless defined($boolean);
   my $list_name = 'digest-users';
   my $list = _get_resource($list_name);
   
   if ($op eq 'set') {
       if ( boolean_value($boolean) ) {
           _list_manip_res($list, ['add' => "!$email",'delete'=> $email], prepend => 1) 
               unless ( _list_manip_res($list,['exists' => "!$email"]) );
       } 
       else {
           _list_manip_res($list, ['add' => $email,'delete'=>"!$email"], prepend => 1)
               unless ( _list_manip_res($list,['exists' => $email]) );
       }
       return 1;
   } 
   else {
       $list->refresh();
       return 0 if $list->match($email);
       return 1;
   }
}

my %EDITABLE_LISTS =  %{$config{lists}};

=item get_lists

Returns a hash of lists and descriptions that this user can modify.

=cut

sub get_lists {
  my ($self) = @_;
  return %EDITABLE_LISTS if wantarray;
  return [ keys %EDITABLE_LISTS ];
}


=item get_items($list_id)

Returns the list contents for C<$list_id>. In an array context returns an array of
lines, in a scalar an array ref.

=cut

# TODO this will be faster with file based lists, but will it be the same for
# dbi lists?
sub get_items {
  my ($self,$list_name) = @_;
  my $list_id = $EDITABLE_LISTS{$list_name};
  die "Invalid List '$list_name'" unless $list_id;
  my $mlist = _get_resource($list_id);
  my $list;
  if (UNIVERSAL::isa($mlist,'PureMessage::MultiList')) {
      $list = $mlist->find_list(key => $self->email, edit_mode => 1);
  } else {
      $list = $mlist;
  }

  die "Invalid List '$list_name'" unless $list;
  my $data = $list->data();
  
  my @items;
  $data->enumerate_items( sub { my $item = shift; push(@items,$item)});
  return @items if wantarray;
  return \@items;
}

=item set_items($list_id,\@items)

Set list C<$list_id> to contain @items.

=cut


sub set_items {
  my ($self,$list_name,$items) = @_;
  my $list_id = $EDITABLE_LISTS{$list_name};
  my $mlist = _get_resource($list_id);
  my $list;
  if (UNIVERSAL::isa($mlist,'PureMessage::MultiList')) {
      $list = $mlist->find_list(key => $self->email, edit_mode => 1);
  } else {
      $list = $mlist;
  }
  die "Invalid List '$list_name'" unless $list_id;
  return _list_manip_res($list,['clear',add => $items],edit_mode => 1);
}



sub available_attributes {
  my $self = shift;
  return keys %ATTRIBUTES;
}

=item set_attribute($name[,$value])

Gets or sets the attibute.

=cut

sub set_attribute {
  my ($self,$attribute,$value) = @_;
  my $code = $ATTRIBUTES{$attribute};
  return  0 unless ref($code) eq 'CODE';
  &$code($self->email,'set',$attribute,$value);
  return 1;
}

sub get_attribute {
  my ($self,$attribute) = @_;
  my $code = $ATTRIBUTES{$attribute};
  return 0 unless ref($code) eq 'CODE';
  return &$code($self->email,'get',$attribute);
}
=item state([$state]) 

Get or set the internal state of this object. This allows queried
data to be re/stored in an external session object if need be.

=cut


sub state {
  my ($self,$state) = @_; 
  if ($state) {
    # then save this state
    $self->{state} = $state;
  } else {
    return $self->{state};
  }
}

=item language()

Get the language attribute for the user.

=cut

sub language() {
    my ($email,$op,$attrib,$value) = @_;
    my $prefs = _get_resource('enduser-preferences');
    if ($op eq 'set') {
	$prefs->language($email, $value);
	$prefs->save;
        _forget_resource('enduser-preferences');
	return $value;
    }
    my $lang = $prefs->language($email);
    _forget_resource('enduser-preferences');
    return $lang;
}

1;
