package PureMessage::RPC::Response::Enduser;
use PureMessage::Service;
use PureMessage::Enduser::Config qw( %config );
use PureMessage::Enduser::User;
use PureMessage::Enduser::Features;
use PureMessage::Manager::Config;
use PureMessage::RPC::Response;
use PureMessage::Quarantine;
use PureMessage::Manager::Lang qw(text mtext);
use PureMessage::Enduser::Session;
use Data::Dumper;

use strict;
use warnings;
our  @ISA = ('PureMessage::RPC::Response');

=head1 NAME

PureMessage::Enduser::Response::Enduser - Enduser RPC Interface

=head1 SYNOPSIS

  This package should not be used directly.

=head1 DESCRIPTION

This package implements the RPC interface for the enduser functionality.

=over

=cut


# check to see if we are allowed to load or if this is disabled
sub _unavailable {
  my $self = shift;
  unless($config{enabled}) {
    # admin has turned off access to this interface
    die $self->raise_detailed_error('unavailable',errorstr=>$config{unavailable_banner}) . "\n";
  }
}


sub log_error {
  my $self = shift;
  if (ref($self->{_request})) {
      $self->{_request}->log_error(@_);
  } else {
    # logging is not enabled TODO log somewhere else?
  }
}

sub _logger {
  my $self = shift;
  return ref($self->{_request}) ? $self->{_request}->_logger() : undef;
}

sub log {
  my $self = shift;
  if (ref($self->{_request})) {
      $self->{_request}->log(@_);
  } else {
    # TODO this needs to work better!
    # logging is not enabled TODO log somewhere else?
  }
}

sub ping {
    my $self = shift;
    my %data = ();
    $self->_unavailable();
    $data{ping} = 'pong';
    $self->dispatch(\%data);
    $self->log('ping');
}

=item RPC auth($email,$pass)

 Attempts to authenticate the passed user. Returns a boolean indicating
 the success.

=cut

sub auth {
    my $self = shift;
    my ($email,$pass) = @_;
    my %data; # data to dispatch
    my $junk;
    my $session;
    $self->_unavailable();
    if (!$email) {
        $self->raise_error('error_rpc_invalid_email');
        return 0;
    }
    my $user;
    eval {
      $user = authorize PureMessage::Enduser::User($email,$pass,logger => $self->_logger());
      die "\n" unless $user; # not much benefit in giving an error
    };
    if ($@) {
      # authentication didn't work, we have to fix
      my $err = $@;
      $self->log(mtext('error_authenticating_user',$email));
      die $self->raise_detailed_error('error_authenticating_user',email => $email, result => $err) . "\n";
    }

    $session = $user->session(create => 1);

    $junk = $session->param(email => $user->email());
    $data{session} = $session->id();
    $self->log(mtext('authenticated_user',$email,$data{session}));

    # make sure that the session is valid and all, otherwise die
    eval { $session->authorize($user->email()) };

    if ($@ or !$session->valid()) {
      $self->log_error(mtext('error_creating_session',$email,$data{session},$session->error()));
      die $self->raise_detailed_error('error_creating_session') . "\n";
    }

    # TODO any session specific operations and defaults that need to be set

    $junk = $self->dispatch(\%data);
    return $session->id(); # very very important or this will get flagged as an error by the handler
}

sub _rebuild_session {
   my $self = shift;
   my $session_id = shift;
   $self->_unavailable();
   my $session = PureMessage::Enduser::Session->restore($session_id);
   die $self->raise_detailed_error('rpc_invalid_session') . "\n" unless $session;
}

sub _get_user {
  my $self = shift;
  my $session = shift; # not a session_id an actual session object
  unless(ref($session)) {
    $self->log_error('error_rpc_invalid_user_in_session','');
    die $self->raise_detailed_error('error_rpc_invalid_user_in_session',email=>'') . "\n";
  }
  my $email = $session->param('email');
  unless ($email) {
    $self->log_error('error_rpc_invalid_user_in_session',$email,$session->id());
    die $self->raise_detailed_error('error_rpc_invalid_user_in_session',email=>$email,session=>$session->id()) . "\n";
  }
  my $data = $session->param('user_data');
  my $user = new PureMessage::Enduser::User($email,logger => $self->_logger());
  $user->restore();
  if(!$user->session() or ($user->session()->id() ne $session->id())) {
    # then for some reason the session_id that this user has stored in it's
    # datastore doesn't match the one that has been presented as
    # authentication. Since the one in the datastore is the true one and only
    # one that should be used we signal an invalid user error.
    #
    $self->log_error('error_rpc_invalid_user_in_session',email=> $email,$session->id());
    die $self->raise_detailed_error('error_rpc_invalid_user_in_session',email=> $email, session=>$session->id()) . "\n";
  }
  return $user;
}

sub _save_user {
  my ($self,$session,$user) = @_;
  $user->store();
}

=item RPC email_credentials($email)

Cause an email with login credentials to be emailed to user.
If successful the mail address credentials were sent to will
be returned.

No authentication at all done on this call.


=cut

sub email_credentials {
  my $self = shift;
  my $email = shift;
  $self->_unavailable();
  $self->log(mtext('email_credentials',$email));
  die $self->raise_detailed_error('authorize_invalid_identity',email=>$email) . "\n" unless $email;

  my $user = new PureMessage::Enduser::User($email,logger => $self->_logger());
  my $result;
  eval {
    $result = $user->send_notification();
  };

  if($@) {
    $self->raise_detailed_error('rpc_error_sending_notification', result=>$@);
    $self->log_error(mtext('rpc_error_sending_notification_log',$email,$@));
    return 0;
  } elsif(!$result) {
    $self->log_error(mtext('rpc_error_sending_notification_log',$email,$@));
    $self->raise_detailed_error('authorize_invalid_identity',email=>$email);
    return 0;
  } else {
    $self->dispatch( $email );
  }
  return 1;
}


=item RPC get_features($session_id)

Returns all the features as a hash

=cut

sub get_features {
  my $self = shift;
  my $session_id = shift;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  # TODO there is a potential to return different features for different
  # users here
  my $features = PureMessage::Enduser::Features::get_all();
  $self->dispatch($features);
  return $session;
}

=item RPC get_user($session_id)

Returns all the known information about this user

=cut

sub get_user {
  my $self = shift;
  my $session_id = shift;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my %result = ();
  $result{email} = $user->email();
  $self->dispatch(\%result);
  return $session;
}

=item RPC init_qview ($session_id,@view)

Initializes the query for the requested quarantine views. If no
view is passed all possible views are initialized.

Returns a row of hashes. Each hash has the keys C<view> and
C<row_count> which contains the count of items for the view.

=cut

sub init_qview {
  my ($self,$session_id,@view) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my @results;

  if (my $err = $user->is_query_too_soon(@view)) {
    $self->raise_detailed_error('quarantine_query_too_soon',errorstr=>$err);
    die $err . "\n";
  }
  # by default don't query all available views, only
  # the ones that are used regularly
  @view = qw( visible delete) unless scalar(@view);
  eval {
    @results = $user->initialize_qview(@view);
  };
  if ($@) {
    die $self->raise_detailed_error('quarantine_query_failed', result => "$@") . "\n";
  }
  $self->log(mtext('quarantine_view_initialized',$user->email,$session_id));
  $self->dispatch(@results);
  $self->_save_user($session,$user);
}

=item RPC get_qview ($session_id,@view)

Returns a row of hashes. Each hash has the keys C<view> and
C<row_count> which contains the count of items for the view.

=cut

sub get_qview {
  my ($self,$session_id,@view) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my @results = $user->get_qview(@view);
  $self->log(mtext('quarantine_qview',$user->email,$session_id));
  $self->dispatch(@results);
  $self->_save_user($session,$user);
  return \@results;
}

=item RPC get_qrows($session,$view,$start,$end[,@sort])

Will retrieve rows $start to $end of the requested quararantine view.
Valid quarantine views are: visible, hidden. This will attempt to
use cached values from the previous C<initialize_qview> call.

You can optionally pass the names of quarantine fields to order the
returned rows. Prefix a field name with a C<-> for descending.

Returns a row of hashes that contain the fields and their values.

=cut


sub get_qrows {
  my ($self,$session_id,$view,$start,$end,@sort) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $rows = $user->get_qrows($view,$start,$end,@sort);
  $self->dispatch(@$rows);
  $self->_save_user($session,$user);
}

# should be used instead of raise_error
# expects the $name to be a valid mtext identifier, the value of each
# hash entry is passed as arguments to the mtext function.
sub raise_detailed_error {
  my ($self,$name,@other) = @_;
  my $i=0;
  my $text = mtext($name,grep { $i++ % 2 } @other);
  $self->log("dispatched error: " . $text);
  $self->raise_error($text);
  return $text;
}

=item RPC get_msg_meta(@ids)

Retreive the meta data for the passed message qid's.

Returns a row of hashes that contain the fields and their values.

=cut

sub getmsg_meta {
  my ($self,$session_id,@ids) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $rows = $user->get_msg_meta(@ids);
  $self->dispatch(@$rows);
}

=item RPC getmsg_body($id)

Will return an html sanitized (but not html encoded) text string of the
messages body content.  The length of the data returned is configurable
by the admin.

=cut

sub getmsg_body {
  my ($self,$session_id,$id) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $text = $user->get_msg_body($id);
  my $meta = $user->get_msg_meta($id);
  $self->log(mtext('quarantine_msg_body_view',$user,$session_id));
  if (defined($text)) {
    my $result;
    $result = $meta->[0] if $meta;
    $result->{body} = $text;
    $self->dispatch( $result );
  } else {
    $self->raise_error('rpc_no_body_content',$id);
  }
}


=item RPC approve_msg(@ids)

Approve the messages that match the passed qid's.

Returns a hash where each key is a passed qid and the
value is true if this operation was successful for that
quid.

=cut

sub approve_msg {
  my ($self,$session_id,@ids) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $result = $user->approve_msg(@ids);
  $self->log(mtext('approve_msg',$user->email,$session_id,join(', ',@ids)));
  $self->dispatch($result);
  $self->_save_user($session,$user);
}

=item RPC delete_msg(@ids)

Delete the messages that match the passed qid's.

Returns a hash where each key is a passed qid and the
value is true if this operation was successful for that
quid.

=cut

sub delete_msg {
  my ($self,$session_id,@ids) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $result = $user->delete_msg(@ids);
  $self->log(mtext('delete_msg',$user->email,$session_id,join(', ',@ids)));
  $self->dispatch($result);
  $self->_save_user($session,$user);
}

=item RPC undelete_msg(@ids)

Undelete the messages that match the passed qid's.

Returns a hash where each key is a passed qid and the
value is true if this operation was successful for that
quid.

=cut

sub undelete_msg {
  my ($self,$session_id,@ids) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $result = $user->undelete_msg(@ids);
  $self->log(mtext('undelete_msg',$user->email,$session_id,join(', ',@ids)));
  $self->dispatch($result);
  $self->_save_user($session,$user);
}

=item RPC get_attribute([$name])

=item RPC set_attribute([$name => $value, .. ])

Set or get the value of attribute $name.

If $name is undef then get_attribute will return the value of
all available attributes.

Set attribute will return the value that the attribute is set to
after the attribute was set. This allows the UI to notice when
an attribute value is read only.



=cut

sub set_attribute {
  my ($self,$session_id,%attributes) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my %data;
  # TODO what happens if a nonexistant attribute is set?
  foreach my $name (keys %attributes) {
    $self->log(mtext('set_attribute',$user->email,$session_id,$name,$attributes{$name}));
    eval {
      $user->set_attribute($name,$attributes{$name});
    };
    if ($@) {
      die $self->raise_detailed_error('set_attribute_error',attribute => $name, errorstr=>$@) . "\n";
    } else {
      $data{$name} = $user->get_attribute($name);
    }
  }
  $self->dispatch(\%data);
  $self->_save_user($session,$user);
}



sub get_attribute {
  my ($self,$session_id,$name) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my %data;
  if ($name) {
    $data{$name} = $user->get_attribute($name);
  } else {
    # get them all
    foreach my $attrib ($user->available_attributes()) {
      $data{$attrib} = $user->get_attribute($attrib);
    }
  }
  $self->dispatch(\%data);
}


=item RPC get_lists($session_id)

Returns the available lists

=cut


sub get_lists {
  my ($self,$session_id) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my %data = $user->get_lists;
  $self->dispatch(\%data);
}

=item RPC get_items($session_id,$list_id)

Returns the items in the requested list as an array.

=cut


sub get_items {
  my ($self,$session_id,$list_id) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);
  my $data = $user->get_items($list_id);
  foreach my $item (@$data) {
     $self->dispatch([$item]);
  }
  $self->dispatch([]) unless scalar(@$data); # return an empty array
  $self->_save_user($session,$user);
}

=item RPC set_items($session_id,$list_id,@items)

Sets the items in the requested list to the passed array.

=cut


sub set_items {
  my ($self,$session_id,$list_id,@items) = @_;
  my $session = $self->_rebuild_session($session_id);
  my $user = $self->_get_user($session);

  die $self->raise_detailed_error('excessive_list_items',limit=>$config{limit}->{max_list_items}) . "\n"
    if scalar(@items) > $config{limit}->{max_list_items};

  my $data = $user->set_items($list_id,\@items);
  $self->log(mtext('set_items',$user->email,$session_id,$list_id));
  $self->dispatch($data);
  $self->_save_user($session,$user);
}

=item get_settings

Returns various useful settings that effect the UI

=cut

sub get_settings {
  my ($self) = @_;
  my %info;
  $self->_unavailable();
  $info{max_vacation} = $config{limit}->{vacation_expiry}; # in seconds
  $info{query_wait} = $config{limit}->{query_wait}; # in seconds
  $info{quarantine_expiry_days} = $config{quarantine_expiry_days} || -1; # in days
  $self->dispatch(\%info);
}





=pod

=back

=cut


=item RPC get_lang()

Returns the language choice for the user.

=cut

sub get_lang {
    my ($self, $session_id) = @_;
    my $session = $self->_rebuild_session($session_id);
    my $user = $self->_get_user($session);
    my $lang = $user->get_lang();
    $self->dispatch($lang);
}

=item RPC set_lang()

Sets the language choice for the user.

=cut

sub set_lang {
    my ($self, $session_id, $lang) = @_;
    my $session = $self->_rebuild_session($session_id);
    my $user = $self->_get_user($session);
    my $ret_value = $user->set_lang($lang);
    return $ret_value;
}


1;
